/*
  Part of the Processing Fullscreen API

  Copyright (c) 2006-08 Hansi Raber

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation; either
  version 3 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
package metasurface;

import java.awt.Component;
import java.awt.Window;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.io.File;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;

import javax.swing.JFileChooser;
import javax.swing.filechooser.FileFilter;

public class MetaSurfaceHelper {
	static Preferences prefs = Preferences.userNodeForPackage( MetaSurfaceHelper.class ); 
	
	/**
	 * Shows a load dialog
	 * this dialog will always remember the last directory it was used with
	 * 
	 * @param key The save key for the preferences
	 * @param file currentFile the currently open file, this can be null
	 * @param fallback If currentfile is null and there is no directory saved in the preferences this will be used. 
	 * @param parent Some parent componet this dialog will behave modal to 
	 * @param type A string containg the allowed file types. (".ext1,.ext2,.ext3", or "" for any, or "dir" for choosing directories)
	 * @return null when the user cancels, the file otherwise 
	 */
	public static File showOpenDialog( String key, File currentFile, File fallback, Component parent, String type ){
		fallback = fallback==null? new File( System.getProperty( "user.dir" ) ) : fallback;
		File chooserFile = currentFile != null? currentFile:new File( prefs.get( key , fallback.getAbsolutePath() ) );
		
		JFileChooser chooser = new JFileChooser( chooserFile );
		if( type.equals( "dir" ) ){
			chooser.setFileSelectionMode( JFileChooser.DIRECTORIES_ONLY ); 
		}
		else if( type.equals( "" ) ){
			chooser.setFileSelectionMode( JFileChooser.FILES_ONLY ); 
		}
		else{
			final String description = type.replaceAll( "/,/", ", " );
			final String extensions[] = type.split( "/,/" ); 
			chooser.setFileSelectionMode( JFileChooser.FILES_ONLY ); 
			chooser.setFileFilter( new FileFilter(){
				@Override
				public boolean accept(File f) {
					for( String ext : extensions )
						if( f.getName().endsWith( ext ) )
							return true; 
						
					return false; 
				}
				
				@Override
				public String getDescription() {
					return description; 
				}
				
			} ); 
		}
		

		int result = chooser.showOpenDialog( parent );
		
		if( result == JFileChooser.APPROVE_OPTION ){
			File file = chooser.getSelectedFile(); 
			prefs.put( "loadDialog." + key, file.getAbsolutePath() ); 
			try {
				prefs.sync();
			} catch (BackingStoreException e) {
				e.printStackTrace();
			} 
			
			return file; 
		}
		else{
			return null; 
		}
	}
	
	/**
	 * Shows a save dialog
	 * @param key The save key for the preferences
	 * @param file currentFile the currently open file, this can be null
	 * @param fallback If currentfile is null and there is no directory saved in the preferences this will be used. 
	 * @param parent Some parent componet this dialog will behave modal to 
	 * @param type A string containg the allowed file types. (".ext1,.ext2,.ext3", or "" for any, or "dir" for choosing directories)
	 * @return null when the user cancels, the file otherwise 
	 */
	public static File showSaveDialog( String key, File currentFile, File fallback, Component parent, String type ){
		fallback = fallback==null? new File( System.getProperty( "user.dir" ) ) : fallback;
		File chooserFile = currentFile != null? currentFile:new File( prefs.get( key , fallback.getAbsolutePath() ) );  
		boolean forceExtension = false; 
		final String extensions[] = type.split( "," );
		
		JFileChooser chooser = new JFileChooser( chooserFile ); 
		if( type.equals( "dir" ) ){
			chooser.setFileSelectionMode( JFileChooser.DIRECTORIES_ONLY ); 
		}
		else if( type.equals( "" ) ){
			chooser.setFileSelectionMode( JFileChooser.FILES_ONLY ); 
		}
		else{
			final String description = type.replaceAll( "/,/", ", " );
			forceExtension = true; 
			chooser.setFileSelectionMode( JFileChooser.FILES_ONLY ); 
			chooser.setFileFilter( new FileFilter(){
				@Override
				public boolean accept( File f ) {
					return hasExtension( f.getName(), extensions ); 
				}
				
				@Override
				public String getDescription() {
					return description; 
				}
				
			} ); 
		}
		
		int result = chooser.showSaveDialog( parent ); 
		
		if( result == JFileChooser.APPROVE_OPTION ){
			File file = chooser.getSelectedFile(); 
			prefs.put( "loadDialog." + key, file.getParent() ); 
			try {
				prefs.sync();
			} catch (BackingStoreException e) {
				e.printStackTrace();
			}
			
			if( !file.exists() && forceExtension && !hasExtension( file.getName(), extensions ) ){
				file = new File( file.getParent(), file.getName() + extensions[0] ); 
			}
			
			return file; 
		}
		else{
			return null; 
		}
	}
	
	/**
	 * Makes sure a window remembers it's size till the next startup
	 */
	public static void rememberSize( final Window window, final String key, int defaultWidth, int defaultHeight ){
		int w = prefs.getInt( key + ".w", defaultWidth );
		int h = prefs.getInt( key + ".h", defaultHeight ); 
		
		window.setSize( w, h );
		
		window.addComponentListener( new ComponentAdapter(){
			@Override
			public void componentResized(ComponentEvent e) {
				prefs.putInt( key + ".w", window.getWidth()  ); 
				prefs.putInt( key + ".h", window.getHeight()  ); 
				try {
					prefs.sync();
				} catch (BackingStoreException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				} 
			}
		} ); 
	}

	/**
	 * Makes sure a window remembers it's size till the next startup
	 */
	public static void rememberLocation( final Window window, final String key, int defaultX, int defaultY ){
		int x = prefs.getInt( key + ".x", defaultX );
		int y = prefs.getInt( key + ".y", defaultY ); 
		
		window.setLocation( x, y );
		
		window.addComponentListener( new ComponentAdapter(){
			@Override
			public void componentMoved(ComponentEvent e) {
				prefs.putInt( key + ".x", window.getLocation().x  ); 
				prefs.putInt( key + ".y", window.getLocation().y  ); 
			}
		} ); 
	}

	/**
	 * Checks if a file has a certain extension
	 */
	private static boolean hasExtension( String fileName, String[] extensions ){
		for( String ext : extensions )
			if( fileName.endsWith( ext ) ) 
				return true; 
		
		return false; 
	}
}
