/*
  Part of the Processing Fullscreen API

  Copyright (c) 2006-08 Hansi Raber

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation; either
  version 3 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
package metasurface;

import java.awt.Dimension;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.Hashtable;

import processing.core.PApplet;
import processing.core.PFont;
import processing.video.MovieMaker;

public class Surface extends PApplet {
	private static final long serialVersionUID = 1L;
	
	MetaSurface dad; 
	Surface me; 
	
	Hashtable <Integer, Entry> entries = new Hashtable<Integer, Entry>();  
	Entry dragEntry = null; 
	Entry activeEntry = null; 
	
	PFont font; 
	int id = 0; 
	
	float pointerX, pointerY; 
	float[] values; 
	
	int pointsX[] = new int[10000]; 
	int pointsY[] = new int[10000]; 
	int frames[] = new int[10000]; 
	int recordPointer = 0; 
	int playbackPointer = 0; 
	int startFrame; 
	 
	
	// stores how "important" each node is (between 0 and 1) 
	
	RegisteredMethods listeners = new RegisteredMethods();
	MovieMaker movieMaker; 
	
	public Surface( MetaSurface dad ){
		this.dad = dad;
		
		values = dad.values;
		
		setPreferredSize( new Dimension( 500, 400 ) );
		setMaximumSize( getPreferredSize() ); 
		setMinimumSize( getPreferredSize() ); 
	}
	
	float dragOffX, dragOffY;
	private boolean recording;
	private boolean playing;
	
	private boolean render; 
	private int renderCount;
	private File renderFile;
	
	public void setup(){
		size( 500, 400 ); 
		me = this; 

		dad.sketch.registerPre( this );
		dad.sketch.registerPost( this ); 
		
		try {
			font = new PFont( Surface.class.getResourceAsStream( "img/Silkscreen-8.vlw" ) );
		} catch (IOException e) {
			System.out.println( "Faild to load font!" ); 
			e.printStackTrace();
			System.exit( 1 ); 
		} 
		frameRate( 20 );
		noLoop();
	}
	
	public void draw(){
		background( 255 ); 
		fill( 0 );
		
		textFont( font );
		textSize( 8 ); 
		
		Entry hover = dragEntry == null?which():dragEntry; 
		
		synchronized( entries ){
			for( Entry entry : entries.values() ){
				if( activeEntry == entry || dragEntry == entry ){
					fill( 255, 0, 0 ); 
					if( hover == entry ) fill( 150, 0, 0 ); 
				}
				else if( hover == entry ){
					fill( 0 ); 
				}
				else{
					fill( 150 ); 
				}
				entry.draw(); 
			}
		}
		
		cursor( hover == null? ARROW:HAND ); 
		
		if( dragEntry != null ){
			dragEntry.x = mouseX - dragOffX; 
			dragEntry.y = mouseY - dragOffY; 
		}
		
		if( mousePressed && dragEntry == null ){
			pointerX = mouseX; 
			pointerY = mouseY; 
			
			calculateValues();
			dad.updateSliders(); 
		}
		
		fill( 255, 0, 0, 150 );
		noStroke(); 
		ellipse( pointerX, pointerY, 10, 10 ); 
		
		fill( 0, 30 );
		for( int i = 0; i < recordPointer; i++ ){
			ellipse( pointsX[i], pointsY[i], 6, 6 ); 
		}
	}
	
	private void drawValues( float x, float y, float values[] ){
		fill( 0, 100 ); 
		for( int i = 0; i < values.length; i++ ){
			
			String text = String.format( "%-10s: %3.3f", dad.params[i], values[i] ); 
			text( text, x, y + i*10 ); 
		}
	}
	
	public void mousePressed(){
		dragEntry = which(); 
		activeEntry = dragEntry; 
		dad.setActiveEntry( activeEntry ); 
		
		if( dragEntry != null ){
			dragOffX = mouseX - dragEntry.x; 
			dragOffY = mouseY - dragEntry.y; 
		}
		
		redraw();
	}
	
	public void mouseReleased(){
		dragEntry = null;
		redraw();
	}
	
	@Override
	public void mouseMoved(){
		redraw(); 
	}
	
	@Override
	public void mouseDragged() {
		redraw();
	}
	
	/**
	 * Which entry the mouse is over
	 */
	private Entry which(){
		synchronized( entries ){
			for( Entry entry : entries.values() ){
				if( entry.over() ){
					return entry; 
				}
			}
		}
		
		return null; 
	}
	
	public synchronized void addEntry( String name, float values[] ){
		if( name.trim().equals( "" ) || name.equals( "new" ) || name.equals( "unknown" )  || name.matches( "NEW-[0-9]+" ) ){
			name = "NEW-" + (id+1);  
		}
		
		synchronized( entries ){
			id ++; 
			entries.put( id, new Entry( id, name, values ) );
		}
		
		activeEntry = entries.get( id ); 
		dad.setActiveEntry( activeEntry ); 
	}
	
	public void removeEntry( Entry entry ){
		if( entry == null ){
			return; 
		}
		
		synchronized( entries ){
			entries.remove( entry.id ); 
			dragEntry = null; 
			activeEntry = null; 
		}
	}
	
	public Entry getEntry( String name, float values ){
		return entries.get( name ); 
	}
	
	public class Entry{
		float values[];
		String name; 
		float x, y; 
		float width, height;
		int id; 
		
		public Entry( int id, String name, float values[] ){
			this.id = id;
			setName( name ); 
			this.values = values; 
		}
		
		public void setName( String name ){
			this.name = name; 
			
			width = textWidth( name );
			height = textAscent() + textDescent(); 
			redraw();
		}
		
		public void setValues( float values[] ){
			System.arraycopy( values, 0, this.values, 0, values.length ); 
			redraw();
		}
		
		public float[] getValues(){
			return values; 
		}
		
		public boolean over(){
			return mouseX >= x && mouseX <= x + width && mouseY >= y && mouseY <= y + 10*(values.length+1) + 5;  
		}
		
		public void draw(){
			x = min( me.width - width, max( 0, x ) );
			y = min( me.height - height, max( 0, y ) ); 
			text( name, x, y + height );
			
			drawValues( x, y + 15, values ); 
		}
	}
	
	
	/**
	 * Calcuate the current average of all the sliders with some crazy formulae 
	 */
	private void calculateValues(){
		// No values?
		synchronized( entries ){
			if( entries.size() == 0 ){
				 
			}
			else{
				// Calculate a weighted average for each of the values... 
				
				float newValues[] = new float[values.length]; 
				float distances[] = new float[values.length];
				
				float dx, dy, d; 
				for( Entry entry : entries.values() ){
					for( int i = 0; i < entry.values.length; i++ ){
						// Distance
						dx = entry.x - pointerX; 
						dy = entry.y - pointerY; 
						d = (float) Math.max( .01f, dx*dx + dy*dy ); 
						newValues[i] += entry.values[i]/d; 
						distances[i] += 1/d; 
					}
				}
				
				for( int i = 0; i < newValues.length; i++ ){
					newValues[i] /= distances[i]; 
				}
				
				System.arraycopy( newValues, 0, values, 0, values.length ); 
			}
		}
	}

		
	public void save( File file ) throws FileNotFoundException{
		PrintStream out = new PrintStream( file );
		
		out.println( "MetaSurface-1.0" ); 
		// Write all the entries... 
		out.println( recordPointer ); 
		
		for( int i = 0; i < recordPointer; i++ ){
			out.println( pointsX[i] + "," + pointsY[i] + "," + frames[i] ); 
		}
		
		out.println( "MetaSurface-Set" );
		
		synchronized( entries ){
			for( Entry entry : entries.values() ){
				out.print( entry.id + "%%%" ); 
				out.print( entry.name + "%%%" ); 
				out.print( entry.x + "%%%" ); 
				out.print( entry.y + "%%%" ); 
				for( float val : entry.values ){
					out.print( val + "," ); 
				}
				out.println(); 
			}
		}
		
		out.close(); 		
	}
	
	public void open( InputStream stream ) throws IOException{
		Hashtable <Integer, Entry> newEntries =  new Hashtable<Integer, Entry>(); 
		int maxId = 0; 
		int[] newPointsX = new int[pointsX.length];
		int[] newPointsY = new int[pointsY.length]; 
		int[] newFrames = new int[frames.length]; 
		int newRecordPointer = 0; 
		BufferedReader in = new BufferedReader( new InputStreamReader( stream ) ); 
		String version = in.readLine(); 
		
		if( version.equals( "MetaSurface-1.0" ) ){
			// we're golden
			newRecordPointer = parseInt( in.readLine() ); 
			
			String line;
			
			int count = 0; 
			while( ( line = in.readLine() ) != null && !line.equals( "MetaSurface-Set" ) ){
				String[] parts = line.split( "," ); 
				if( parts.length == 3 ){
					// Now get the floats
					newPointsX[count] = parseInt( parts[0] ); 
					newPointsY[count] = parseInt( parts[1] ); 
					newFrames[count] = parseInt( parts[2] );
					count++;
				}					
				else{
					System.err.println( "Incompatible length" ); 
					return; 
				}
			}
			
			while( ( line = in.readLine() ) != null ){
				String[] parts = line.split( "%%%" ); 
				
				if( parts.length == 5 ){
					// Now get the floats
					String[] floats = parts[4].split( "," );
					float[] vals = new float[values.length]; 
					
					// In case they have different length... warn, but try to merge!
					for( int i = 0; i < floats.length && i < vals.length; i++ ){
						vals[i] = parseFloat( floats[i] ); 
					}
					
					if( vals.length != floats.length ){
						System.err.println( "Huh - the file you're loading seems to be made for a different configuration of the metasurface" ); 
						System.err.println( "I'm trying my best, but this is most likely gonna fuck up. " );
						System.err.println(); 
					}
					
					int id = parseInt( parts[0] );
					maxId = Math.max( id, maxId ); 
					String name = parts[1]; 
					int x = parseInt( parts[2] ); 
					int y= parseInt( parts[3] );
					
					Entry entry = new Entry( id, name,vals ); 
					entry.x = x; 
					entry.y = y; 
					
					newEntries.put( id, entry ); 
				}
			}
			
		}
		
		id = maxId + 1; 
		dad.setActiveEntry( null ); 
		dragEntry = activeEntry = null; 
		
		entries = newEntries; 
		pointsX = newPointsX; 
		pointsY = newPointsY; 
		frames = newFrames; 
		recordPointer = newRecordPointer; 
	}
	
	public void open( File file ) throws FileNotFoundException, IOException{
		open( new FileInputStream( file ) );
	}
	
	public void startRecording(){
		if( recording ) return; 
		listeners.handle( new Object[]{ new MetaSurfaceEvent( dad, "startRecording" ) } );
		dad.setActiveEntry( activeEntry = null ); 
		recordPointer = 0; 
		Arrays.fill( pointsX, 0 ); 
		Arrays.fill( pointsY, 0 ); 
		Arrays.fill( frames, 0 ); 
		playbackPointer = 0;
		startFrame = dad.sketch.frameCount;  
		recording = true;
	}
	
	public void stopRecording(){
		if( !recording ) return; 
		// add a last coordinate!
		pre(); 
		recording = false; 
		listeners.handle( new Object[]{ new MetaSurfaceEvent( dad, "stopRecording" ) } );
	}
	
	public void startPlayback(){
		playbackPointer = 0; 
		playing = true;
		startFrame = dad.sketch.frameCount;
		listeners.handle( new Object[]{ new MetaSurfaceEvent( dad, "start" ) } );
		dad.setActiveEntry( activeEntry = null ); 
	}
	
	public void stopPlayback(){
		playing = false; 
		listeners.handle( new Object[]{ new MetaSurfaceEvent( dad, "stop" ) } );
	}
	
	public boolean isPlaying(){
		return playing; 
	}
	
	public void startRender( File file ){
		// try to delete the file, this saves some headaches
		if( movieMaker == null && renderFile == null ){
			renderFile = file;
			render = true; 
		}
	}
	
	public void stopRender(){
		if( movieMaker != null && renderFile != null){
			render = false;  
		}
	}
	
	/**
	 * Called by processing
	 */
	public void pre(){
		if( recording ){
			if( recordPointer == 0 ){
				// Only the first mouse click triggers the record event 
				if( mousePressed ){
					pointerX = mouseX; 
					pointerY = mouseY; 
					startFrame = dad.sketch.frameCount; 
					pointsX[recordPointer] = (int) pointerX; 
					pointsY[recordPointer] = (int) pointerY;
					frames[recordPointer] = dad.sketch.frameCount - startFrame;
					recordPointer ++; 
				}
			}
			else if( pointsX[recordPointer-1] != pointerX || pointsY[recordPointer-1] != pointerY ){
				pointsX[recordPointer] = (int) pointerX; 
				pointsY[recordPointer] = (int) pointerY;
				frames[recordPointer] = dad.sketch.frameCount - startFrame; 
				recordPointer ++;
			}
		}
		
		if( playing ){
			if( playbackPointer == 0 || (dad.sketch.frameCount - startFrame > frames[playbackPointer] && playbackPointer < recordPointer ) ){
				pointerX = pointsX[playbackPointer]; 
				pointerY = pointsY[playbackPointer]; 
				calculateValues();
				dad.updateSliders(); 
				redraw(); 
				
				playbackPointer++;
				
				if( playbackPointer >= recordPointer ){
					stopPlayback();  
					dad.stop.setEnabled( false ); 
					dad.play.setEnabled( true ); 
				}
			}
		}
	}

	/**
	 * Called by processing
	 */
	public void post(){
		// Create moviemake
		if( render && movieMaker == null && renderFile != null ){
			renderCount = 0; 
			renderFile.delete();
			listeners.handle( new Object[]{ new MetaSurfaceEvent( dad, "startRender" ) } );			
			movieMaker = new MovieMaker( dad.sketch, dad.sketch.width, dad.sketch.height, renderFile.getAbsolutePath(), dad.getExportFrameRate(), MovieMaker.MOTION_JPEG_B, MovieMaker.BEST );
		}
		
		// Add frame
		if( render && movieMaker != null && renderFile != null ){
			System.out.println( renderFile.getAbsolutePath() + ": frame " +  + renderCount + " @" + dad.sketch.frameRate + "fps" );
			renderCount ++; 
			//dad.sketch.updatePixels();
			//dad.sketch.loadPixels(); 
			movieMaker.addFrame( ); 
		}
		
		// Stop 
		if( !render && movieMaker != null && renderFile != null ){
			movieMaker.finish(); 
			movieMaker = null; 
			PApplet.open( renderFile.getAbsolutePath() ); 
			renderFile = null; 
			movieMaker = null; 
			listeners.handle( new Object[]{ new MetaSurfaceEvent( dad, "stopRender" ) } );			
		}
	}
	
	public void registerListener(Object o) {
		registerWithArgs( listeners, "metaSurfaceEvent", o, new Class[]{ MetaSurfaceEvent.class } ); 
	}
}
